(* 565 Pset 2 *)


Require Import Omega.
(* Omega provides a useful tactic "omega" that solves quantifier 
 * free formula expressed in Presburger arithmetic.  See:
 * https://coq.inria.fr/refman/Reference-Manual024.html
 *)

(*
 * This problem set examines how we might reason about a simple language
 * of arithmetic expressions.
 *)
Module ArithWithConstants.

  (* The following definition closely mirrors a standard BNF grammar for expressions.
   * It defines abstract syntax trees of arithmetic expressions. *)

  Inductive arith : Set :=
  | Const (n : nat)
  | Plus (e1 e2 : arith)
  | Times (e1 e2 : arith).

  (* Here are a few examples of specific expressions. *)
  Example ex1 := Const 42.
  Example ex2 := Plus (Const 1) (Times (Const 2) (Const 3)).

  Fixpoint size (e : arith) : nat.
    (* Write a definition that computes the size of an arithmetic expression
     * This should be equivalent to the size of the expression's abstract syntax tree
     *)
    Admitted.

  Fixpoint depth (e : arith) : nat.
  (* What's the longest path from the root of a syntax tree to a leaf? *)
    Admitted.

  Theorem depth_le_size : forall e, depth e <= size e.
  Proof.
    Admitted.

  
  Fixpoint commuter (e : arith) : arith.
    (* commuter is a function that swaps the order of arguments on all binary operations 
     * Once you define this function, demonstrate through examples that it operates the
     * way you expect. 
    *)
    Admitted.


  Theorem size_commuter : forall e, size (commuter e) = size e.
  Proof.
    Admitted.
     
  Theorem commuter_inverse : forall e, commuter (commuter e) = e.
   Proof.
    Admitted.

End ArithWithConstants.

Module ArithWithVariables.

Require Import String.  
Notation var := string.
Definition var_eq : forall x y : var, {x = y} + {x <> y} := string_dec.

Infix "==v" := var_eq (no associativity, at level 50).

(* We now consider an extension of our language to include variables *)
Inductive arith : Set :=
  | Const (n : nat)
  | Var (x : var) (* <-- this is the new constructor! *)
  | Plus (e1 e2 : arith)
  | Times (e1 e2 : arith).

Fixpoint size (e : arith) : nat.
  (* augment earlier definition *)
  Admitted.

Fixpoint depth (e : arith) : nat.
  (* augment earlier definition *)
  Admitted.
  
Fixpoint substitute (inThis : arith) (replaceThis : var) (withThis : arith) : arith.
  (* Given expression "inThis", replace all occurrences of variable "replacethis" with
   * expression "withThis"
  *)
Admitted.

Theorem substitute_depth : forall replaceThis withThis inThis,
    depth (substitute inThis replaceThis withThis) <= depth inThis + depth withThis.
Proof.
  (* Substitution never shrinks the size of expressions *)
  Admitted.
      
Fixpoint constantFold (e : arith) : arith.
  (* Given an arithmetic expression whose arguments are all constants, replace the
   * expression with the value of the expression.  E.g., replace "Plus 2 3" with
   * "Const (2 + 3)"
  *)
Admitted.

Theorem size_constantFold : forall e, size (constantFold e) <= size e.
Proof.
  Admitted.

End ArithWithVariables.
