(* 565 Pset 2 *)


Require Import Omega.
Require Import Pset2.
(* Omega provides a useful tactic "omega" that solves quantifier 
 * free formula expressed in Presburger arithmetic.  See:
 * https://coq.inria.fr/refman/Reference-Manual024.html
 *)

(*
 * This problem set examines how we might reason about a simple language
 * of arithmetic expressions.
 *)
Module ArithWithConstants.
Import Pset2.ArithWithConstants.

  (* The following definition closely mirrors a standard BNF grammar for expressions.
   * It defines abstract syntax trees of arithmetic expressions. *)

  Print Assumptions size.
  Print Assumptions depth.

  Theorem depth_le_size : forall e, depth e <= size e.
  Proof.
   apply depth_le_size.
   Qed. 
  
  Print Assumptions depth_le_size.

  Print Assumptions commuter.

  Theorem size_commuter : forall e, size (commuter e) = size e.
  Proof.
    apply size_commuter.
   Qed.
 
  Print Assumptions size_commuter.
     
  Theorem commuter_inverse : forall e, commuter (commuter e) = e.
   Proof.
    apply commuter_inverse.
   Qed.

  Print Assumptions commuter_inverse.

End ArithWithConstants.

Module ArithWithVariables.
Import Pset2.ArithWithVariables.

Require Import String.  
Notation var := string.
Definition var_eq : forall x y : var, {x = y} + {x <> y} := string_dec.

Infix "==v" := var_eq (no associativity, at level 50).

(* We now consider an extension of our language to include variables *)

Print Assumptions size.
Print Assumptions depth.
Print Assumptions substitute.

Theorem substitute_depth : forall replaceThis withThis inThis,
    depth (substitute inThis replaceThis withThis) <= depth inThis + depth withThis.
Proof.
  (* Substitution never shrinks the size of expressions *)
  apply substitute_depth.
   Qed.

Print Assumptions substitute_depth.

Print Assumptions constantFold.
      
Theorem size_constantFold : forall e, size (constantFold e) <= size e.
Proof.
  apply size_constantFold.
   Qed.

Print Assumptions size_constantFold.

End ArithWithVariables.
